// tslint:disable:no-console
import { Result } from "../../src/util/result";
import { cSnRectang, ParamsSectionRectang } from "../../src/section/section_rectang";
import { MessageCode } from "../../src/util/message";

let paramSection: ParamsSectionRectang;
let sect: cSnRectang;

describe("Section paramétrée rectangulaire : ", () => {
    beforeEach(() => {
        paramSection = new ParamsSectionRectang(0.8, // tirant d'eau
            2.5, // largeur de fond
            40, //  Ks=Strickler
            10,  //  Q=Débit
            0.001, //  If=pente du fond
            1e-10, // précision
            1, // YB= hauteur de berge
        );

        sect = new cSnRectang(paramSection);
        sect.newtonMaxIter = 5;
    });

    describe("non convergence de la méthode de Newton :", () => {
        it("hauteur critique", () => {
            const r: Result = sect.Calc("Yc");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
        });

        it("hauteur normale", () => {
            const r: Result = sect.Calc("Yn");
            // expect(r.log.messages.length).toEqual(2);
            expect(r.ok).toBeFalsy("le calcul devrait avoir échoué");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[1].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HNORMALE);
        });

        it("hauteur normale, pente nulle", () => {
            sect.prms.If.v = 0;
            const r: Result = sect.Calc("Yn");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_PENTE_NEG_NULLE_HNORMALE_INF);
        });

        it("hauteur normale, pente négative", () => {
            sect.prms.If.v = -0.001;
            const r: Result = sect.Calc("Yn");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_PENTE_NEG_NULLE_HNORMALE_INF);
        });

        it("hauteur fluviale, Y < Yc", () => {
            const r: Result = sect.Calc("Yf");
            // expect(r.log.messages.length).toEqual(3);
            expect(r.ok).toBeFalsy("le calcul devrait avoir échoué");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[1].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[2].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HFLU);
        });

        it("hauteur fluviale, Y > Yc", () => {
            sect.prms.Y.v = 2;
            const r: Result = sect.Calc("Yf");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
        });

        it("hauteur torrentielle, Y < Yc", () => {
            const r: Result = sect.Calc("Yt");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
        });

        it("hauteur torrentielle, Y > Yc", () => {
            sect.prms.Y.v = 2;
            const r: Result = sect.Calc("Yt");
            //            expect(r.log.messages.length).toEqual(2);
            expect(r.ok).toBeFalsy("le calcul devrait avoir échoué");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[1].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HTOR);
        });

        it("hauteur conjuguée, Froude < 1", () => {
            sect.prms.Y.v = 2;
            const r: Result = sect.Calc("Yco");
            // console.log(r.log.toString());
            // expect(r.log.messages.length).toEqual(3);
            expect(r.ok).toBeFalsy("le calcul devrait avoir échoué");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[1].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HTOR);
            // expect(r.log.messages[2].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCONJUG);
        });

        it("hauteur conjuguée, Froude > 1", () => {
            const r: Result = sect.Calc("Yco");
            // expect(r.log.messages.length).toEqual(4);
            expect(r.ok).toBeFalsy("le calcul devrait avoir échoué");
            expect(r.log.messages.length).toEqual(1);
            expect(r.log.messages[0].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[1].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
            // expect(r.log.messages[2].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HFLU);
            // expect(r.log.messages[3].code).toEqual(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCONJUG);
        });
    });
});
