// import { XOR, BoolIdentity, Debug, Result, ResultCode, UndefinedError } from "./base";
import { BoolIdentity, Debug, XOR } from "./base";
import { Nub } from "./nub";
import { ParamDefinition, ParamDomain, ParamDomainValue } from "./param";
import { Interval } from "./util/interval";
import { Message, MessageCode } from "./util/message";
import { Pair } from "./util/pair";
import { Result } from "./util/result";

class SearchInterval extends Interval {
    private _step: number;

    private _dicho: Dichotomie;

    private _targets: Pair;

    constructor(d: Dichotomie, min: number, max: number, s: number) {
        super(min, max);
        if (s === 0) {
            const e = new Message(MessageCode.ERROR_DICHO_NULL_STEP);
            throw e;
        }
        this._step = s;
        this._dicho = d;
    }

    public reverse() {
        this._step = -this._step;
    }

    public growStep(k: number) {
        if (k === 0) {
            const e = new Message(MessageCode.ERROR_DICHO_INVALID_STEP_GROWTH);
            throw e;
        }
        this._step *= k;
    }

    get step() {
        return this._step;
    }

    get targets() {
        if (this._targets === undefined) {
            this._targets = new Pair(undefined, undefined);
        }
        return this._targets;
    }

    /**
     * get target value for lower bound
     */
    get targetLower() {
        this.updateTargets();
        return this.targets.val1;
    }

    /**
     * get target value for upper bound
     */
    get targetUpper() {
        this.updateTargets();
        return this.targets.val2;
    }

    public next() {
        if (this._step > 0) {
            this._val1 = this._val2;
            this._val2 += this._step;
            this.targets.setValues(this.targets.val2, undefined);
        } else {
            this._val2 = this._val1;
            this._val1 += this._step;
            this.targets.setValues(undefined, this.targets.val1);
        }
    }

    public hasTargetValue(targ: number) {
        this.updateTargets();
        return this.targets.intervalHasValue(targ);
    }

    public toString(): string {
        return super.toString() + " step=" + this._step;
    }

    private updateTargets() {
        let t1 = this.targets.val1;
        if (t1 === undefined) {
            t1 = this._dicho.CalculX(this._val1).vCalc;
        }

        let t2 = this.targets.val2;
        if (t2 === undefined) {
            t2 = this._dicho.CalculX(this._val2).vCalc;
        }
        this.targets.setValues(t1, t2);
    }

}

/**
 * calcul par dichotomie
 * principe : y=f(x1,x2,...), on connait une valeur de y,
 * on cherche par ex le x2 correspondant, les autres xi étant fixés.
 * la méthode Equation() calcule analytiquement y=f(xi)
 */
// tslint:disable-next-line:max-classes-per-file
export class Dichotomie extends Debug {
    /**
     * définition de la variable de la fonction
     */
    private _paramX: ParamDefinition;

    /**
     * nombre d'étapes de recherche de l'intervalle de départ
     */
    private _startIntervalMaxSteps = 100;

    /**
     * nom du paramètre calculé analytiquement par Equation()
     */
    private analyticalSymbol: string;

    /**
     * Construction de la classe.
     * @param nub Noeud de calcul contenant la méthode de calcul Equation
     * @param sVarCalc Nom de la variable à calculer
     * @param targetSymbol nom du paramètre calculé analytiquement par Equation()
     */
    constructor(private nub: Nub, private sVarCalc: string, dbg: boolean = false, targetSymbol?: string) {
        super(dbg);
        this._paramX = this.nub.getParameter(this.sVarCalc);
        if (targetSymbol === undefined) {
            this.analyticalSymbol = this.nub.getFirstAnalyticalParameter().symbol;
        } else {
            this.analyticalSymbol = targetSymbol;
        }
    }

    /**
     * Valeur inconnue à rechercher
     */
    private get vX(): number {
        return this._paramX.v;
    }

    private set vX(vCalc: number) {
        this._paramX.v = vCalc;
    }

    /**
     * étapes de recherche de l'intervalle de départ
     */
    get startIntervalMaxSteps() {
        return this._startIntervalMaxSteps;
    }

    set startIntervalMaxSteps(n: number) {
        this._startIntervalMaxSteps = n;
    }

    public CalculX(x: number): Result {
        this.vX = x;
        return this.Calcul();
    }

    /**
     * trouve la valeur x pour y=f(x) avec un y donné
     * @param rTarget valeur de y
     * @param rTol tolérance pour l'arrêt de la recherche
     * @param rInit valeur initiale approximative de x
     */
    public Dichotomie(rTarget: number, rTol: number, rInit: number): Result {
        // console.log("-----");
        // for (let x = 0; x <= 1; x += 0.1)
        //     console.log(this.CalculX(x).vCalc);
        // console.log("-----");

        // recherche de l'intervalle de départ

        this.debug(
            "Dichotomie : valeur cible de " + this.analyticalSymbol + "=" + rTarget
            + ", valeur initiale de " + this.sVarCalc + "=" + rInit
        );

        const r = this.getStartInterval(rTarget, rInit);
        if (r.ok) {
            const interv: SearchInterval = r.intSearch;
            // Dichotomie
            return this.dichoSearch(rTarget, rTol, interv.min, interv.max, interv.targetLower, interv.targetUpper);
        } else {
            return new Result(r.res);
        }
    }

    /**
     * Calcul de l'équation analytique.
     * @note Wrapper vers this.nub.Equation pour simplifier le code.
     * On utilise la première variable du tableau des variables pouvant être calculée analytiquement
     * Il faudra s'assurer que cette première variable correspond à la méthode de calcul la plus rapide
     */
    private Calcul(): Result {
        const r: Result = this.nub.Calc(this.analyticalSymbol);
        this.debug(
            "dicho : Calcul(vX=" + this.sVarCalc + "=" + this.vX + ") -> " +
            this.analyticalSymbol + "=" + r.vCalc);
        return r;
    }

    /**
     * Détermine si une fonction est croissante ou décroissante.
     * @param x point auquel on calcul la dérivée
     * @param dom domaine de définition de la variable
     */
    private isIncreasingFunction(x: number, dom: Interval): boolean {
        const epsilon = 1e-8;
        const bounds = new Interval(x - epsilon, x + epsilon);
        bounds.setBounds(bounds.intersect(dom)); // au cas où l'on sorte du domaine de la variable de la fonction

        const y1 = this.CalculX(bounds.min).vCalc;
        const y2 = this.CalculX(bounds.max).vCalc;
        return y2 > y1;
    }

    /**
     * recherche l'intervalle contenant la valeur cible
     * @param rTarget valeur cible
     * @param intSearch intervalle de départ
     * @param intMax intervalle maximum (domaine de définition de la variable)
     */
    private searchTarget(rTarget: number, intSearch: SearchInterval, intMax: Interval) {
        this.debug("searchTarget : Target " + rTarget + " dans " + intSearch.toString());

        let n = 0;
        let ok: boolean = false;
        do {
            const inters: Interval = intSearch.intersect(intMax);
            if (inters.length === 0) {
                break;
            }
            intSearch.setBounds(inters);
            if (intSearch.hasTargetValue(rTarget)) {
                ok = true;
                break;
            }
            intSearch.growStep(2);
            intSearch.next();
        } while (n++ < this._startIntervalMaxSteps);

        return { ok, intSearch };
    }

    /**
     * Détermine l'intervalle de recherche initial
     * @param rTarget valeur cible de la fonction
     * @param rInit  valeur initiale de la variable
     */
    private getStartInterval(rTarget: number, rInit: number): any {
        this.debug("intervalle de départ");

        const prmDom: ParamDomain = this._paramX.getDomain();
        const step = 0.001;

        const min: number = prmDom.minValue;
        const max: number = prmDom.maxValue;
        if (prmDom.domain === ParamDomainValue.NOT_NULL) {
            if (rInit === 0) {
                rInit = 1e-8;
            }
        }

        const intMax: Interval = new Interval(min, max);
        try {
            intMax.checkValue(rInit);
        } catch (m) {
            return { ok: false, res: m };
        }

        // initialisation de l'intervalle de recherche
        let intSearch: SearchInterval = new SearchInterval(this, rInit - step / 2, rInit + step / 2, step);
        // au cas où l'on sorte du domaine de la variable de la fonction
        intSearch.setBounds(intSearch.intersect(intMax));

        // sens de variation de la fonction
        const inc = this.isIncreasingFunction(rInit, intMax);

        const initTarget = this.CalculX(rInit).vCalc;
        // if (initTarget > rTarget && inc || initTarget < rTarget && !inc)
        if (BoolIdentity(initTarget > rTarget, inc)) {
            intSearch.reverse();
            // par ex, la fonction est croissante et la valeur initiale
            // de la variable a une image par la fonction > valeur cible
        }

        // on cherche dans une première direction

        let a = this.searchTarget(rTarget, intSearch, intMax);
        if (a.ok) {
            return a;
        }

        // il se peut que la fonction ne soit pas monotone et qu'il faille chercher dans l'autre direction

        const oldStepSign = intSearch.step > 0 ? 1 : -1;
        intSearch = new SearchInterval(this, rInit + step / 2, rInit + step, step * -oldStepSign);
        // au cas où l'on sorte du domaine de la variable de la fonction
        intSearch.setBounds(intSearch.intersect(intMax));

        a = this.searchTarget(rTarget, intSearch, intMax);
        if (a.ok) {
            return a;
        }

        // gestion de l'erreur

        // la valeur cible de la fonction est elle trouvable ?
        let res: Message;
        let errDomain = false;
        switch (prmDom.domain) {
            case ParamDomainValue.INTERVAL:
                const si: SearchInterval = new SearchInterval(this, intMax.min, intMax.max, 1);
                errDomain = !si.hasTargetValue(rTarget);
                break;

            case ParamDomainValue.POS:
            case ParamDomainValue.POS_NULL:
                const y = this.CalculX(1e-8).vCalc;
                errDomain = inc && (rTarget < y);
                break;

            case ParamDomainValue.ANY:
                break;

            default:
                throw new Error("unsupported parameter domain value");
        }

        if (errDomain) {
            res = new Message(MessageCode.ERROR_DICHO_INIT_DOMAIN);
            res.extraVar.targetSymbol = this.analyticalSymbol; // symbole de la variable calculée par la fonction
            res.extraVar.targetValue = rTarget; // valeur cible pour la fonction
            // intervalle de valeurs pour la variable d'entrée de la fonction
            res.extraVar.variableInterval = intMax.toString();
            res.extraVar.variableSymbol = this._paramX.symbol; // symbole de la variable d'entrée de la fonction
        } else {
            // if (intSearch.step < 0)
            if (BoolIdentity(initTarget > rTarget, inc)) {
                res = new Message(MessageCode.ERROR_DICHO_INITVALUE_HIGH);
            } else {
                res = new Message(MessageCode.ERROR_DICHO_INITVALUE_LOW);
            }
            res.extraVar.variableSymbol = this._paramX.symbol; // symbole de la variable de la fonction
            res.extraVar.variableInitValue = rInit; // valeur initiale de la variable
            res.extraVar.targetSymbol = this.analyticalSymbol; // symbole de la variable calculée par la fonction
            res.extraVar.targetValue = rTarget; // valeur cible pour la fonction
            res.extraVar.initTarget = initTarget; // valeur de la fonction pour la valeur initiale de la variable
        }

        return { ok: false, res };
    }

    /**
     * effectue récursivement la recherche dichotomique
     * @param Xmin valeur min de l'intervalle
     * @param Xmax valeur max de l'intervalle
     * @param Y valeur cible de la fonction
     * @param Ymin valeur de la fonction pour Xmin
     * @param Ymax valeur de la fonction pour Xmax
     */
    // tslint:disable-next-line:variable-name
    private dichoSearch(Ytarg: number, rTol: number, Xmin: number, Xmax: number, Ymin: number, Ymax: number): Result {
        this.debug("dichoSearch : yTarget " + Ytarg + " X " + Xmin + "->" + Xmax + " tol " + rTol);

        // tslint:disable-next-line:variable-name
        const Xmid = (Xmin + Xmax) / 2;

        if (Math.abs(Xmin - Xmax) <= rTol) {
            return new Result(Xmid);
        }

        // tslint:disable-next-line:variable-name
        const Ymid = this.CalculX(Xmid).vCalc;
        if (Ymin < Ymax) {
            // fonction croissante

            if (Ytarg > Ymid) {
                return this.dichoSearch(Ytarg, rTol, Xmid, Xmax, Ymid, Ymax);
            } // intervalle supérieur
            return this.dichoSearch(Ytarg, rTol, Xmin, Xmid, Ymin, Ymid); // intervalle inférieur
        } else {
            // fonction décroissante

            if (Ytarg > Ymid) {
                return this.dichoSearch(Ytarg, rTol, Xmin, Xmid, Ymin, Ymid);
            } // intervalle inférieur
            return this.dichoSearch(Ytarg, rTol, Xmid, Xmax, Ymid, Ymax); // intervalle supérieur
        }
    }

}
