import { Debug } from "../base"
import { acSection, cParamsCanal } from "./section_type";
import { acNewton } from "./newton";
import { Result } from "../util/result";
import { MessageCode, Message } from "../util/message";

/**
 * Calcul de la hauteur critique
 */
export class cHautCritique extends acNewton {
        /**
         * Section sur laquuelle porte le calcul
         */
        private Sn: acSection;

        /**
         * Constructeur de la classe
         * @param Sn Section sur laquelle on fait le calcul
         */
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Sn = Sn.clone();
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcFn(rX: number): Result {
                const rS: Result = this.Sn.Calc("S", rX);
                if (!rS.ok)
                        return rS;

                // Calcul de la fonction
                // if (this.Sn.Calc("S", rX) != 0)
                if (rS.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));

                // return (Math.pow(this.Sn.prms.Q.v, 2) * this.Sn.Calc("B", rX) / Math.pow(this.Sn.Calc("S", rX), 3) / cParamsCanal.G - 1);
                const rB: Result = this.Sn.Calc("B", rX);
                if (!rB.ok)
                        return rB;

                const rS2: Result = this.Sn.Calc("S", rX);
                if (!rS2.ok)
                        return rS2;

                const v = (Math.pow(this.Sn.prms.Q.v, 2) * rB.vCalc / Math.pow(rS2.vCalc, 3) / cParamsCanal.G - 1);
                return new Result(v);

                // return Infinity;
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcDer(rX: number): Result {
                // let S = this.Sn.Calc("S");
                const rS: Result = this.Sn.Calc("S");
                if (!rS.ok)
                        return rS;
                const S = rS.vCalc;
                // if (S != 0) {
                if (S == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));

                // let B = this.Sn.Calc("B");
                const rB: Result = this.Sn.Calc("B");
                if (!rB.ok)
                        return rB;
                const B = rB.vCalc;

                const rDB: Result = this.Sn.Calc("dB");
                if (!rDB.ok)
                        return rDB;

                // L'initialisation à partir de rX a été faite lors de l'appel à CalcFn
                // let Der = (this.Sn.Calc("dB") * S - 3 * B * B);
                let Der = (rDB.vCalc * S - 3 * B * B);
                const v = Math.pow(this.Sn.prms.Q.v, 2) / cParamsCanal.G * Der / Math.pow(S, 4);
                return new Result(v);
                // }

                // return Infinity;
        }
}

/**
 * Calcul de la hauteur normale
 */
export class cHautNormale extends acNewton {
        /**
       * Section sur laquuelle porte le calcul
       */
        private Sn: acSection;

        /**
         * Débit connu
         */
        private Q: number;

        /**
         * Coefficient de Strickler
         */
        private Ks: number;

        /**
         * Pente du fond
         */
        private If: number;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Sn = Sn;
                this.Q = Sn.prms.Q.v;
                this.Ks = Sn.prms.Ks.v;
                this.If = Sn.prms.If.v;
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcFn(rX: number): Result {
                // Calcul de la fonction
                // return (this.Q - this.Ks * Math.pow(this.Sn.Calc("R", rX), 2 / 3) * this.Sn.Calc("S", rX) * Math.sqrt(this.If));
                const rR: Result = this.Sn.Calc("R", rX);
                if (!rR.ok)
                        return rR;

                const rS: Result = this.Sn.Calc("S", rX);
                if (!rS.ok)
                        return rS;

                const v = (this.Q - this.Ks * Math.pow(rR.vCalc, 2 / 3) * rS.vCalc * Math.sqrt(this.If));
                return new Result(v);
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcDer(rX: number): Result {
                const rDR: Result = this.Sn.Calc("dR");
                if (!rDR.ok)
                        return rDR;

                const rR: Result = this.Sn.Calc("R");
                if (!rR.ok)
                        return rR;

                const rS: Result = this.Sn.Calc("S");
                if (!rS.ok)
                        return rS;

                // L'initialisation a été faite lors de l'appel à CalcFn

                // let Der = 2 / 3 * this.Sn.Calc("dR") * Math.pow(this.Sn.Calc("R"), -1 / 3) * this.Sn.Calc("S");
                let Der = 2 / 3 * rDR.vCalc * Math.pow(rR.vCalc, -1 / 3) * rS.vCalc;

                const rR2: Result = this.Sn.Calc("R");
                if (!rR2.ok)
                        return rR2;

                const rB: Result = this.Sn.Calc("B");
                if (!rB.ok)
                        return rB;

                // Der = Der + Math.pow(this.Sn.Calc("R"), 2 / 3) * this.Sn.Calc("B");
                Der = Der + Math.pow(rR2.vCalc, 2 / 3) * rB.vCalc;

                Der = Der * -this.Ks * Math.sqrt(this.If);
                return new Result(Der);
        }
}

/**
 * Calcul de la hauteur correspondante (charge égale)
 */
export class cHautCorrespondante extends acNewton {
        /**
         * Tirant d'eau connu
         */
        private Y: number;

        /** 
         * 1/S^2 associé au tirant d'eau connu
         */
        private _rS2: Result;

        /**
         * Section contenant les données de la section avec la hauteur à calculer
         */
        private Sn: acSection;

        /**
         * Constante de gravité
         */
        private rQ2G: number;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Y = Sn.prms.Y.v;
                //                this.rS2 = Math.pow(Sn.Calc("S"), -2);
                this.Sn = Sn;
                this.rS2;  // pour initialiser la valeur
                this.rQ2G = Math.pow(Sn.prms.Q.v, 2) / (2 * cParamsCanal.G);
        }

        private get rS2(): Result {
                if (this._rS2 == undefined) {
                        const rS = this.Sn.Calc("S");
                        if (rS.ok) {
                                const v = Math.pow(rS.vCalc, -2);
                                this._rS2 = new Result(v);
                        }
                        else
                                this._rS2 = rS;
                }
                return this._rS2;
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcFn(rX: number): Result {
                if (!this.rS2.ok)
                        return this.rS2;

                const rS: Result = this.Sn.Calc("S", rX);
                if (!rS.ok)
                        return rS;

                // Calcul de la fonction
                // return this.Y - rX + (this.rS2 - Math.pow(this.Sn.Calc("S", rX), -2)) * this.rQ2G;
                const v = this.Y - rX + (this.rS2.vCalc - Math.pow(rS.vCalc, -2)) * this.rQ2G;
                return new Result(v);
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on protected function cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcDer(rX: number): Result {
                // let S = this.Sn.Calc("S");
                const rS: Result = this.Sn.Calc("S");
                if (!rS.ok)
                        return rS;
                const S = rS.vCalc;

                // L'initialisation a été faite lors de l'appel à CalcFn
                // if (S != 0)
                if (S == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));

                const rB: Result = this.Sn.Calc("B");
                if (!rB.ok)
                        return rB;

                // return -1 + 2 * this.rQ2G * this.Sn.Calc("B") / Math.pow(S, 3);
                const v = -1 + 2 * this.rQ2G * rB.vCalc / Math.pow(S, 3);
                return new Result(v);

                // return Infinity;
        }
}

/**
 * Calcul de la hauteur conjuguée (Impulsion égale)
 */
export class cHautConjuguee extends acNewton {
        /** Section contenant les données de la section avec la hauteur à calculer */
        private Sn: acSection;

        /** Carré du débit */
        private rQ2: number;

        /** Surface hydraulique associée au tirant d'eau connu */
        private rS: Result;

        /** SYg associée au tirant d'eau connu */
        private rSYg: Result;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        constructor(oSn: acSection, maxIter: number, dbg: boolean = false) {
                super(oSn.prms, maxIter, dbg);
                this.rQ2 = Math.pow(oSn.prms.Q.v, 2);
                this.Sn = oSn;
                this.rS = oSn.Calc("S");
                this.rSYg = oSn.Calc("SYg");
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcFn(rX: number) {
                if (!this.rS.ok)
                        return this.rS;

                if (!this.rSYg.ok)
                        return this.rSYg;

                const rS2: Result = this.Sn.Calc("S", rX);
                if (!rS2.ok)
                        return rS2;

                // Réinitialisation des paramètres hydrauliques de this.Sn avec l'appel this.Sn.Calc("S',rX)
                // if (this.rS > 0 && this.Sn.Calc("S", rX) > 0) {
                if (this.rS.vCalc > 0 && rS2.vCalc > 0) {
                        // let Fn = this.rQ2 * (1 / this.rS - 1 / this.Sn.Calc("S"));
                        const rS3: Result = this.Sn.Calc("S");
                        if (!rS3.ok)
                                return rS3;

                        let Fn = this.rQ2 * (1 / this.rS.vCalc - 1 / rS3.vCalc);

                        let rSYg = this.Sn.Calc("SYg");
                        if (!rSYg.ok)
                                return rSYg;

                        // return Fn + cParamsCanal.G * (this.rSYg - this.Sn.Calc("SYg"));
                        const v = Fn + cParamsCanal.G * (this.rSYg.vCalc - rSYg.vCalc);
                        return new Result(v);
                }

                return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));
                // return -Infinity;
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        CalcDer(rX: number): Result {
                if (!this.rS.ok)
                        return this.rS;

                //                let S = this.Sn.Calc("S");
                const rS2: Result = this.Sn.Calc("S");
                if (!rS2.ok)
                        return rS2;
                const S = rS2.vCalc;

                // L'initialisation a été faite lors de l'appel à CalcFn
                // if (this.rS > 0 && S > 0) {
                if (this.rS.vCalc > 0 && S > 0) {
                        const rDS: Result = this.Sn.Calc("dS");
                        if (!rDS.ok)
                                return rDS;

                        // let Der = this.rQ2 * this.Sn.Calc("dS") * Math.pow(S, -2);
                        let Der = this.rQ2 * rDS.vCalc * Math.pow(S, -2);

                        const rDYG: Result = this.Sn.Calc("dSYg", rX);
                        if (!rDYG.ok)
                                return rDYG;

                        // return Der - cParamsCanal.G * this.Sn.Calc("dSYg", rX);
                        const v = Der - cParamsCanal.G * rDYG.vCalc;
                        return new Result(v);
                }

                return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));
                // return -Infinity;
        }
}
