import { ComputeNodeType, ParamDefinition, ParamDomainValue, ParamCalculability } from "../param";
import { acSection, ParamsSection } from "./section_type";
import { Result } from "../util/result";

export class ParamsSectionTrapez extends ParamsSection {
        private _LargeurFond: ParamDefinition; // Largeur au fond
        private _Fruit: ParamDefinition; // Fruit des berges

        constructor(rLargeurFond: number, rFruit: number, rY: number, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number,
                nodeType: ComputeNodeType = ComputeNodeType.SectionTrapeze,
                defaultNodeType: ComputeNodeType = ComputeNodeType.SectionParametree) {
                super(rY, undefined, rKs, rQ, rIf, rPrec, rYB, defaultNodeType);
                this._LargeurFond = new ParamDefinition(nodeType, 'LargeurFond', ParamDomainValue.POS_NULL, rLargeurFond);
                this._Fruit = new ParamDefinition(nodeType, 'Fruit', ParamDomainValue.POS_NULL, rFruit);

                this.addParamDefinition(this._LargeurFond);
                this.addParamDefinition(this._Fruit);
        }

        /**
        * Largeur au fond
        */
        get LargeurFond(): ParamDefinition {
                return this._LargeurFond;
        }

        /**
        * Fruit des berges
        */
        get Fruit(): ParamDefinition {
                return this._Fruit;
        }
}


/**
 * Calculs de la section trapézoïdale
 */
export class cSnTrapez extends acSection {
        constructor(prms: ParamsSectionTrapez, dbg: boolean = false) {
                super(prms, dbg);
        }

        protected setParametersCalculability() {
                super.setParametersCalculability();
                this.prms.LargeurFond.calculability = ParamCalculability.DICHO;
                this.prms.Fruit.calculability = ParamCalculability.DICHO;
        }

        get prms(): ParamsSectionTrapez {
                return <ParamsSectionTrapez>this._prms;
        }

        private isDebordement(): boolean {
                return this.prms.Y.v > this.prms.YB.v;
        }

        protected Calc_B(): Result {
                if (this.isDebordement()) {
                        const v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.YB.v;
                        return new Result(v);
                }

                const v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul du périmètre mouillé
         * @return Périmètre mouillé (m)
         */
        protected Calc_P(): Result {
                if (this.isDebordement()) {
                        // return this.CalcGeo("P") + super.Calc_P_Debordement(this.prms.Y.v - this.prms.YB.v);

                        const rGeoP: Result = this.CalcGeo("P");
                        if (!rGeoP.ok)
                                return rGeoP;

                        const rPDeb: Result = super.Calc_P_Debordement(this.prms.Y.v - this.prms.YB.v);
                        if (!rPDeb.ok)
                                return rPDeb;

                        const v = rGeoP.vCalc + rPDeb.vCalc;
                        return new Result(v);
                }

                const v = this.prms.LargeurFond.v + 2 * Math.sqrt(1 + Math.pow(this.prms.Fruit.v, 2)) * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul de la surface mouillée
         * @return Surface mouillée (m2)
         */
        protected Calc_S(): Result {
                if (this.isDebordement()) {
                        // return this.CalcGeo("S") + super.Calc_S_Debordement(this.prms.Y.v - this.prms.YB.v);
                        const rGeoS: Result = this.CalcGeo("S");
                        if (!rGeoS.ok)
                                return rGeoS;

                        const rSDeb: Result = super.Calc_S_Debordement(this.prms.Y.v - this.prms.YB.v);
                        if (!rSDeb.ok)
                                return rSDeb;

                        const v = rGeoS.vCalc + rSDeb.vCalc;
                        return new Result(v);
                }

                const v = this.prms.Y.v * (this.prms.LargeurFond.v + this.prms.Fruit.v * this.prms.Y.v);
                return new Result(v);
        }

        /**
         * Calcul de dérivée de la surface hydraulique par rapport au tirant d'eau.
         * @return dS
         */
        protected Calc_dS(): Result {
                if (this.isDebordement())
                        return super.Calc_dS();

                const v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected Calc_dP(): Result {
                if (this.isDebordement())
                        return super.Calc_dP_Debordement();

                const v = 2 * Math.sqrt(1 + this.prms.Fruit.v * this.prms.Fruit.v);
                return new Result(v);
        }

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
         * @return dB
         */
        protected Calc_dB(): Result {
                if (this.isDebordement())
                        return super.Calc_dB_Debordement();

                const v = 2 * this.prms.LargeurFond.v * this.prms.Fruit.v;
                return new Result(v);
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): Result {
                const v = (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3) * Math.pow(this.prms.Y.v, 2);
                return new Result(v);
        }

        /**
         * Calcul de la dérivée de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): Result {
                var SYg = this.prms.Fruit.v / 3 * Math.pow(this.prms.Y.v, 2);
                SYg += (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3) * 2 * this.prms.Y.v;
                return new Result(SYg);
        }
}