import { MessageCode, Message } from "../util/message";
import { ComputeNodeType, ComputeNode, ParamDefinition, ParamDomainValue, ParamCalculability, ParamsEquation } from "../param";
import { cHautCritique, cHautNormale, cHautCorrespondante, cHautConjuguee } from "./hauteur";
import { Result } from "../util/result";

/**
 * Gestion des Paramètres du canal (hors section)
 */
export abstract class cParamsCanal extends ParamsEquation {
        public static readonly G: number = 9.81;/// Constante de gravité

        private _Ks: ParamDefinition; // Strickler
        private _Q: ParamDefinition; // Débit
        private _If: ParamDefinition;  // Pente du fond
        private _Prec: ParamDefinition;  // Précision de calcul et d'affichage
        private _iPrec: ParamDefinition;  // Précision en nombre de décimales
        private _YB: ParamDefinition;  // Hauteur de berge

        constructor(nodeType: ComputeNodeType, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number) {
                super();
                this._Ks = new ParamDefinition(nodeType, 'Ks', ParamDomainValue.POS, rKs);
                this._Q = new ParamDefinition(nodeType, 'Q', ParamDomainValue.POS_NULL, rQ);
                this._If = new ParamDefinition(nodeType, 'If', ParamDomainValue.ANY, rIf);
                this._Prec = new ParamDefinition(nodeType, 'Prec', ParamDomainValue.POS, rPrec);
                this._iPrec = new ParamDefinition(nodeType, 'iPrec', ParamDomainValue.ANY, Math.round(-Math.log(rPrec) / Math.log(10)));
                this._YB = new ParamDefinition(nodeType, 'YB', ParamDomainValue.POS, rYB);

                this.addParamDefinition(this._Ks);
                this.addParamDefinition(this._Q);
                this.addParamDefinition(this._If);
                this.addParamDefinition(this._Prec);
                this.addParamDefinition(this._iPrec);
                this.addParamDefinition(this._YB);
        }

        /**
         * Strickler
         */
        get Ks(): ParamDefinition {
                return this._Ks;
        }

        /**
         * Débit
         */
        get Q(): ParamDefinition {
                return this._Q;
        }

        /**
         * Pente du fond
         */
        get If(): ParamDefinition {
                return this._If;
        }

        /**
         * Précision de calcul et d'affichage
         */
        get Prec(): ParamDefinition {
                return this._Prec;
        }

        /**
         * Précision en nombre de décimales
         */
        get iPrec(): ParamDefinition {
                return this._iPrec;
        }

        /**
         * Hauteur de berge
         */
        get YB(): ParamDefinition {
                return this._YB;
        }
}


export abstract class ParamsSection extends cParamsCanal {
        private _Y: ParamDefinition;          // Tirant d'eau
        private _LargeurBerge: ParamDefinition; // largeur au débordement

        constructor(rY: number, rLargeurBerge: number, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number,
                nodeType: ComputeNodeType = ComputeNodeType.SectionParametree) {
                super(nodeType, rKs, rQ, rIf, rPrec, rYB);
                this._Y = new ParamDefinition(nodeType, 'Y', ParamDomainValue.POS_NULL, rY);
                this._LargeurBerge = new ParamDefinition(nodeType, 'LargeurBerge', ParamDomainValue.POS_NULL, rLargeurBerge);

                this.addParamDefinition(this._Y);
                this.addParamDefinition(this._LargeurBerge);
        }

        /**
        * Tirant d'eau
        */
        get Y(): ParamDefinition {
                return this._Y;
        }

        /**
        * Largeur au débordement
        */
        get LargeurBerge(): ParamDefinition {
                return this._LargeurBerge;
        }
}

/**
 * Gestion commune pour les différents types de section.
 * Comprend les formules pour la section rectangulaire pour gérer les débordements
 */
export abstract class acSection extends ComputeNode {
        [key: string]: any; // pour pouvoir faire this['methode]();

        private _hautCritique: Result;  // Tirant d'eau critique

        /**
         * true si la section est fermée (on considère alors qu'il existe une fente de Preissmann)
         */
        protected bSnFermee: boolean = false;

        /**
         * Tableau contenant les données dépendantes du tirant d'eau this->rY.
         *
         * Les clés du tableau peuvent être :
         * - S : la surface hydraulique
         * - P : le périmètre hydraulique
         * - R : le rayon hydraulique
         * - B : la largeur au miroir
         * - J : la perte de charge
         * - Fr : le nombre de Froude
         * - dP : la dérivée de P par rapport Y
         * - dR : la dérivée de R par rapport Y
         * - dB : la dérivée de B par rapport Y
         */
        private arCalc: { [key: string]: number } = {};
        protected arCalcGeo: { [key: string]: number } = {}; /// Données ne dépendant pas de la cote de l'eau
        private Y_old: number; /// Mémorisation du tirant d'eau pour calcul intermédiaire
        private Calc_old = {}; /// Mémorisation des données hydrauliques pour calcul intermédiaire

        /**
         * itérations max pour Newton
         */
        private _newtonMaxIter: number = 50;

        /**
         * flag de débuggage pour Newton
         */
        private _newtonDbg: boolean;

        private _indentCalc: number = -1;

        /**
         * Construction de la classe.
         * Calcul des hauteurs normale et critique
         */
        constructor(prms: ParamsSection, dbg: boolean = false) {
                super(prms, dbg);

                // this._v = oP.v;  // copie la référence au tableau
                // this.prms.Y.v = 0;
                // this.prms.LargeurBerge.v = 0;

                // this.CalcGeo("B");  // à ce point, il manque des paramètres pour certaines sections (ex : fruit pour trapèze)

                this._newtonDbg = dbg;
        }

        protected setParametersCalculability() {
                this.prms.Ks.calculability = ParamCalculability.DICHO;
                this.prms.Q.calculability = ParamCalculability.DICHO;
                this.prms.If.calculability = ParamCalculability.DICHO;
                this.prms.Prec.calculability = ParamCalculability.NONE;
                this.prms.iPrec.calculability = ParamCalculability.NONE;
                this.prms.YB.calculability = ParamCalculability.DICHO;
                this.prms.Y.calculability = ParamCalculability.DICHO;
                this.prms.LargeurBerge.calculability = ParamCalculability.DICHO;
        }

        get prms(): ParamsSection {
                return <ParamsSection>this._prms;
        }

        public clone(): acSection {
                return Object.create(this);
        }

        public set newtonMaxIter(n: number) {
                this._newtonMaxIter = n;
        }

        public get HautCritique() {
                return this._hautCritique;
        }

        /**
         * Efface toutes les données calculées pour forcer le recalcul
         * @param bGeo Réinitialise les données de géométrie aussi
         */
        public Reset(bGeo = true) {
                this.debug("reset(" + bGeo + ")");
                this.arCalc = {};
                if (bGeo) {
                        this.arCalcGeo = {};
                        this.bSnFermee = false;
                }
        }

        /**
         * Mémorise les données hydraulique en cours ou les restitue
         * @param bMem true pour mémorisation, false pour restitution
         */
        private Swap(bMem: boolean) {
                if (bMem) {
                        this.debug("save Y=" + this.prms.Y.toString());
                        this.Y_old = this.prms.Y.uncheckedValue;
                        this.Calc_old = this.arCalc;
                }
                else {
                        this.debug("restore Y=" + this.Y_old);
                        this.prms.Y.v = this.Y_old;
                        this.arCalc = this.Calc_old;
                        this.Calc_old = {};
                }
        }

        private calcIndent(): string {
                let res = "";
                for (let i = 0; i < this._indentCalc; i++)
                        res += "  ";
                return res;
        }

        public debug(m: string) {
                super.debug(this.calcIndent() + m);
        }

        /**
         * Calcul des données à la section
         * @param sDonnee Clé de la donnée à calculer (voir this->arCalc)
         * @param rY valeur de Y à utiliser
         * @return la donnée calculée
         */
        public Calc(sDonnee: string, rY: number = undefined): Result {
                this._indentCalc++;
                this.debug("in Calc(" + sDonnee + ", rY=" + rY + ") old " + sDonnee + "=" + this.arCalc[sDonnee]);
                this.debug("this.Y=" + this.prms.Y.toString());

                if (rY != undefined && (!this.prms.Y.isDefined || rY != this.prms.Y.v)) {
                        this.prms.Y.v = rY;
                        // On efface toutes les données dépendantes de Y pour forcer le calcul
                        this.Reset(false);
                }
                if (this.arCalc[sDonnee] == undefined) {
                        // La donnée a besoin d'être calculée
                        switch (sDonnee) {
                                case 'I-J': // Variation linéaire de l'énergie spécifique (I-J) en m/m
                                        // this.arCalc[sDonnee] = this.prms.If.v - this.Calc("J");
                                        var rJ: Result = this.Calc("J");
                                        if (rJ.ok) {
                                                this.arCalc[sDonnee] = this.prms.If.v - rJ.vCalc;
                                                var res = new Result(this.arCalc[sDonnee]);
                                        }
                                        else
                                                res = rJ;
                                        break;

                                default:
                                        let methode = 'Calc_' + sDonnee;

                                        /* !!!!!
                                           si on réunit les 2 lignes suivantes ( this.arCalc[sDonnee] = this[methode](); ),
                                           il arrive que this.arCalc[sDonnee] soit affecté à undefined alors que this[methode]() renvoie
                                           une valeur bien définie
                                           !!!!!
                                         */
                                        res = this[methode]();
                                        if (res.ok)
                                                this.arCalc[sDonnee] = res.vCalc;
                                        break;
                        }
                        this.debug("Calc(" + sDonnee + ") resultat -> " + this.arCalc[sDonnee]);
                }
                else {
                        this.debug("Calc(" + sDonnee + ") cache= " + this.arCalc[sDonnee]);
                        res = new Result(this.arCalc[sDonnee]);
                }
                this._indentCalc--;

                this.debug('this.Y=' + this.prms.Y.toString());

                // return this.arCalc[sDonnee];
                return res;
        }

        /**
         * Calcul des données uniquement dépendantes de la géométrie de la section
         * @param sDonnee Clé de la donnée à calculer (voir this->arCalcGeo)
         * @return la donnée calculée
         */
        protected CalcGeo(sDonnee: string): Result {
                this._indentCalc++;
                this.debug("in CalcGeo(" + sDonnee + ") old " + sDonnee + "=" + this.arCalcGeo[sDonnee]);
                this.debug("this.Y=" + this.prms.Y.toString());

                // Si la largeur aux berges n'a pas encore été calculée, on commence par ça
                if (sDonnee != "B" && this.arCalcGeo["B"] == undefined)
                        this.CalcGeo("B");

                if (this.arCalcGeo[sDonnee] == undefined) {
                        // La donnée a besoin d'être calculée
                        this.Swap(true); // On mémorise les données hydrauliques en cours
                        this.Reset(false);
                        this.prms.Y.v = this.prms.YB.v;
                        switch (sDonnee) {
                                case "B": // Largeur aux berges
                                        // this.arCalcGeo[sDonnee] = this.Calc_B();
                                        var res: Result = this.Calc_B();
                                        if (res.ok) {
                                                this.arCalcGeo[sDonnee] = res.vCalc;
                                                if (this.arCalcGeo[sDonnee] < this.prms.YB.v / 100) {
                                                        // Section fermée
                                                        this.bSnFermee = true;
                                                        // On propose une fente de Preissmann égale à 1/100 de la hauteur des berges
                                                        this.arCalcGeo[sDonnee] = this.prms.YB.v / 100;
                                                }
                                                this.prms.LargeurBerge.v = this.arCalcGeo[sDonnee];
                                        }
                                        break;

                                default:
                                        let methode = "Calc_" + sDonnee;
                                        // this.arCalcGeo[sDonnee] = this[methode]();
                                        res = this[methode]();
                                        if (res.ok)
                                                this.arCalcGeo[sDonnee] = res.vCalc;
                                        break;
                        }
                        this.Swap(false); // On restitue les données hydrauliques en cours
                        this.debug("CalcGeo(" + sDonnee + ") resultat -> " + this.arCalcGeo[sDonnee]);
                }
                else {
                        this.debug("CalcGeo(" + sDonnee + ") cache= " + this.arCalcGeo[sDonnee]);
                        res = new Result(this.arCalcGeo[sDonnee]);
                }

                // this.debug('this.Y=' + this.prms.Y.toString());

                this._indentCalc--;

                // return this.arCalcGeo[sDonnee];
                return res;
        }

        /**
         * Calcul de la surface hydraulique.
         * @return La surface hydraulique
         */
        protected abstract Calc_S(): Result;


        protected Calc_dS(): Result {
                return this.Calc("B"); // largeur au miroir
        }

        /**
         * Calcul de la surface hydraulique en cas de débordement
         * @param Y hauteur d'eau au delà de la berge
         **/
        protected Calc_S_Debordement(Y: number): Result {
                this.debug('section->CalcS(rY=' + Y + ') LargeurBerge=' + this.CalcGeo("B"));
                // return Y * this.CalcGeo("B");
                let rB: Result = this.CalcGeo("B");
                if (!rB.ok)
                        return rB;

                return new Result(Y * rB.vCalc);
        }

        /**
         * Calcul de la dérivée surface hydraulique en cas de débordement
         * @return La dérivée de la surface hydraulique en cas de débordement
         */
        private Calc_dS_Debordement(): Result {
                return this.CalcGeo("B");
        }

        /**
         * Calcul du périmètre hydraulique.
         * @return Le périmètre hydraulique
         */
        protected abstract Calc_P(): Result;

        /**
         * Calcul du périmètre hydraulique en cas de débordement
         * @param Y hauteur d'eau au dela de la berge
         */
        protected Calc_P_Debordement(Y: number): Result {
                return new Result(2 * Y);
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected abstract Calc_dP(): Result;

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau en cas de débordement
         * @return la dérivée du périmètre hydraulique par rapport au tirant d'eau en cas de débordement
         */
        protected Calc_dP_Debordement(): Result {
                return new Result(2);
        }

        /**
         * Calcul du rayon hydraulique.
         * @return Le rayon hydraulique
         */
        private Calc_R(): Result {
                // let P = this.Calc("P");
                let rP: Result = this.Calc("P");
                if (!rP.ok)
                        return rP;

                // if (P != 0)
                if (rP.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));
                // return this.Calc("S") / P;

                let rS = this.Calc("S");
                if (!rS.ok)
                        return rS;

                return new Result(rS.vCalc / rP.vCalc);

                // return Infinity;
        }

        /**
         * Calcul de dérivée du rayon hydraulique par rapport au tirant d'eau.
         * @return dR
         */
        private Calc_dR(): Result {
                // let P = this.Calc("P");
                const rP: Result = this.Calc("P");

                if (!rP.ok)
                        return rP;

                // if (P != 0)
                if (rP.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));

                //return ((this.Calc("B") * P - this.Calc("S") * this.Calc("dP")) / Math.pow(P, 2));
                const rB: Result = this.Calc("B");
                if (!rB.ok)
                        return rB;

                const rS: Result = this.Calc("S");
                if (!rS.ok)
                        return rS;

                const rDP: Result = this.Calc("dP");
                if (!rDP.ok)
                        return rDP;

                const v = ((rB.vCalc * rP.vCalc - rS.vCalc * rDP.vCalc) / Math.pow(rP.vCalc, 2));
                return new Result(v);

                // return 0;
        }

        /**
         * Calcul de la largeur au miroir.
         * @return La largeur au miroir
         */
        protected abstract Calc_B(): Result;


        /**
           * Calcul de la largeur au miroir en cas de débordement
           * @return La largeur au miroir en cas de débordement
           */
        protected Calc_B_Debordement(): Result {
                // return this.prms.LargeurBerge.v;
                return new Result(this.prms.LargeurBerge.v);
        }

        /**
          * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
          * @return dB
          */
        protected abstract Calc_dB(): Result;

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau en cas de débordement
         * @return la dérivée de la largeur au miroir par rapport au tirant d'eau en cas de débordement
         */
        protected Calc_dB_Debordement(): Result {
                return new Result(0);
        }

        /**
         * Calcul de la perte de charge par la formule de Manning-Strickler.
         * @return La perte de charge
         */
        private Calc_J(): Result {
                // let R = this.Calc("R");
                let rR: Result = this.Calc("R");
                if (!rR.ok)
                        return rR;

                // if (R != 0)
                if (rR.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_RAYON_NUL));

                // return Math.pow(this.Calc("V") / this.prms.Ks.v, 2) / Math.pow(R, 4 / 3);
                const rV: Result = this.Calc("V");
                if (!rV.ok)
                        return rV;

                const v = Math.pow(rV.vCalc / this.prms.Ks.v, 2) / Math.pow(rR.vCalc, 4 / 3);
                return new Result(v);
                //                return Infinity;
        }

        /**
         * Calcul du nombre de Froude.
         * @return Le nombre de Froude
         */
        private Calc_Fr(): Result {
                // let S = this.Calc("S");
                const rS: Result = this.Calc("S");

                if (!rS.ok)
                        return rS;

                // if (S != 0)
                if (rS.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));

                const rB: Result = this.Calc("B");
                if (!rB.ok)
                        return rB;

                // return this.prms.Q.v / S * Math.sqrt(this.Calc("B") / S / cParamsCanal.G);
                const v = this.prms.Q.v / rS.vCalc * Math.sqrt(rB.vCalc / rS.vCalc / cParamsCanal.G);
                return new Result(v);

                // return Infinity;
        }

        /**
         * Calcul de dy/dx
         */
        private Calc_dYdX(Y: number): Result {
                // L'appel à Calc("J') avec Y en paramètre réinitialise toutes les données dépendantes de la ligne d'eau

                // return - (this.prms.If.v - this.Calc("J", Y) / (1 - Math.pow(this.Calc("Fr", Y), 2)));
                const rJ: Result = this.Calc("J", Y);
                if (!rJ.ok)
                        return rJ;

                const rFR: Result = this.Calc("Fr", Y);
                if (!rFR.ok)
                        return rFR;

                const v = - (this.prms.If.v - rJ.vCalc / (1 - Math.pow(rFR.vCalc, 2)));
                return new Result(v);
        }

        /**
         * Calcul de la vitesse moyenne.
         * @return Vitesse moyenne
         */
        private Calc_V(): Result {
                // let S = this.Calc("S");
                let rS: Result = this.Calc("S");
                if (!rS.ok)
                        return rS;

                // if (S != 0)
                if (rS.vCalc == 0)
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));

                // return this.prms.Q.v / S;
                const v = this.prms.Q.v / rS.vCalc;
                return new Result(v);

                // return Infinity;
        }

        /**
         * Calcul de la charge spécifique.
         * @return Charge spécifique
         */
        private Calc_Hs(): Result {
                // return this.prms.Y.v + Math.pow(this.Calc("V"), 2) / (2 * cParamsCanal.G);

                let rV: Result = this.Calc("V");
                if (!rV.ok)
                        return rV;

                const v = this.prms.Y.v + Math.pow(rV.vCalc, 2) / (2 * cParamsCanal.G);
                return new Result(v);
        }

        /**
         * Calcul de la charge spécifique critique.
         * @return Charge spécifique critique
         */
        private Calc_Hsc(): Result {
                let res: Result;

                this.Swap(true); // On mémorise les données hydrauliques en cours

                // On calcule la charge avec la hauteur critique
                // var Hsc = this.Calc("Hs", this.CalcGeo("Yc"));
                const rYC: Result = this.CalcGeo("Yc");
                if (!rYC.ok)
                        res = rYC;
                else
                        res = this.Calc("Hs", rYC.vCalc);

                // On restitue les données initiales
                this.Swap(false);

                // return Hsc;
                return res;
        }

        /**
         * Calcul du tirant d'eau critique.
         * @return tirant d'eau critique
         */
        private Calc_Yc(): Result {
                let hautCritique = new cHautCritique(this, this._newtonMaxIter, this.newtonDbg);
                this._hautCritique = hautCritique.Newton(this.prms.YB.v);

                //                if (this.HautCritique == undefined || !hautCritique.hasConverged()) {
                if (!this._hautCritique.ok) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
                        // this.oLog.add(m);
                        this._hautCritique = new Result(m);
                }
                return this._hautCritique;
        }

        /**
         * Calcul du tirant d'eau normal.
         * @return tirant d'eau normal
         */
        private Calc_Yn(): Result {
                if (this.prms.If.v <= 0) {
                        let m = new Message(MessageCode.ERROR_SECTION_PENTE_NEG_NULLE_HNORMALE_INF);
                        // this.oLog.add(m);
                        // return undefined;
                        return new Result(m);
                }

                const rYC: Result = this.CalcGeo("Yc");
                if (!rYC.ok)
                        return rYC;

                let oHautNormale = new cHautNormale(this, this._newtonMaxIter, this._newtonDbg);
                // let res = oHautNormale.Newton(this.CalcGeo("Yc"));
                let rYN = oHautNormale.Newton(rYC.vCalc);

                // if (res == undefined || !oHautNormale.hasConverged()) {
                if (!rYN.ok) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HNORMALE);
                        // this.oLog.add(m);
                        rYN = new Result(m);
                }

                return rYN;
        }

        /**
         * Calcul du tirant d'eau fluvial.
         * @return tirant d'eau fluvial
         */
        private Calc_Yf(): Result {
                const rGeoYC: Result = this.CalcGeo("Yc");
                if (!rGeoYC.ok)
                        return rGeoYC;

                // if (this.prms.Y.v > this.CalcGeo("Yc"))
                if (this.prms.Y.v > rGeoYC.vCalc)
                        return new Result(this.prms.Y.v);

                const oHautCorrespondante = new cHautCorrespondante(this, this._newtonMaxIter, this._newtonDbg);

                const rYC: Result = this.Calc("Yc");
                if (!rYC.ok)
                        return rYC;
                // let res = oHautCorrespondante.Newton(this.Calc("Yc") * 2);
                let rYF = oHautCorrespondante.Newton(rYC.vCalc * 2);

                // if (res == undefined || !oHautCorrespondante.hasConverged()) {
                if (!rYF.ok) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HFLU);
                        // this.oLog.add(m);
                        rYF = new Result(m);
                }

                return rYF;
        }

        /**
         * Calcul du tirant d'eau torrentiel.
         * @return tirant d'eau torrentiel
         */
        private Calc_Yt(): Result {
                const rGeoYC: Result = this.CalcGeo("Yc");
                if (!rGeoYC.ok)
                        return rGeoYC;

                // if (this.prms.Y.v < this.CalcGeo("Yc"))
                if (this.prms.Y.v < rGeoYC.vCalc)
                        return new Result(this.prms.Y.v);

                var oHautCorrespondante = new cHautCorrespondante(this, this._newtonMaxIter, this._newtonDbg);
                // let res = oHautCorrespondante.Newton(this.CalcGeo("Yc") / 2);
                let rYT = oHautCorrespondante.Newton(rGeoYC.vCalc / 2);
                // if (res == undefined || !oHautCorrespondante.hasConverged()) {
                if (!rYT.ok) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HTOR);
                        // this.oLog.add(m);
                        rYT = new Result(m);
                }

                return rYT;
        }

        /**
         * Calcul du tirant d'eau conjugué.
         * @return tirant d'eau conjugué
         */
        private Calc_Yco(): Result {
                let res: Result;

                this.Swap(true);

                const oHautConj = new cHautConjuguee(this, this._newtonMaxIter, this._newtonDbg);

                const rFR: Result = this.Calc("Fr");
                if (!rFR.ok)
                        res = rFR;
                else {
                        // Choisir une valeur initiale du bon côté de la courbe
                        // if (this.Calc("Fr") < 1) {
                        if (rFR.vCalc < 1) {
                                // Ecoulement fluvial, on cherche la conjuguée à partir du tirant d'eau torrentiel
                                var Y0: Result = this.Calc("Yt");
                        }
                        else {
                                // Ecoulement torrentiel, on cherche la conjuguée à partir du tirant d'eau fluvial
                                Y0 = this.Calc("Yf");
                        }
                        if (!Y0.ok)
                                res = Y0;
                        else {
                                // let Yco = oHautConj.Newton(Y0);
                                const Yco: Result = oHautConj.Newton(Y0.vCalc);

                                // if (Yco == undefined || !oHautConj.hasConverged()) {
                                if (!Yco.ok) {
                                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCONJUG);
                                        // this.oLog.add(m);
                                        res = new Result(m);
                                }
                                else
                                        res = Yco;
                        }
                }

                this.Swap(false);
                return res;
        }

        /**
         * Calcul de la contrainte de cisaillement.
         * @return contrainte de cisaillement
         */
        private Calc_Tau0(): Result {
                const rR: Result = this.Calc("R");
                if (!rR.ok)
                        return rR;

                const rJ: Result = this.Calc("J");
                if (!rJ.ok)
                        return rJ;

                // return 1000 * cParamsCanal.G * this.Calc("R") * this.Calc("J");
                return new Result(1000 * cParamsCanal.G * rR.vCalc * rJ.vCalc);
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): Result {
                // return Math.pow(this.prms.Y.v, 2) * this.CalcGeo("B") / 2;

                const rGeoB: Result = this.CalcGeo("B");
                if (!rGeoB.ok)
                        return rGeoB;

                // return Math.pow(this.prms.Y.v, 2) * this.CalcGeo("B") / 2;
                const v = Math.pow(this.prms.Y.v, 2) * rGeoB.vCalc / 2;
                return new Result(v);
        }

        /**
         * Calcul de la dérivée distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): Result {
                const rGeoB: Result = this.CalcGeo("B");
                if (!rGeoB.ok)
                        return rGeoB;

                // return this.prms.Y.v * this.CalcGeo("B");
                const v = this.prms.Y.v * rGeoB.vCalc;
                return new Result(v);
        }

        /**
         * Calcul de l'impulsion hydraulique.
         * @return Impulsion hydraulique
         */
        private Calc_Imp(): Result {
                const rV: Result = this.Calc("V");
                if (!rV.ok)
                        return rV;

                const rSYG: Result = this.Calc("SYg");
                if (!rSYG.ok)
                        return rSYG;

                // return 1000 * (this.prms.Q.v * this.Calc("V") + cParamsCanal.G * this.Calc("SYg"));
                const v = 1000 * (this.prms.Q.v * rV.vCalc + cParamsCanal.G * rSYG.vCalc);
                return new Result(v);
        }

        /**
         * Calcul de l'angle Alpha entre la surface libre et le fond pour les sections circulaires.
         * @return Angle Alpha pour une section circulaire, 0 sinon.
         */
        protected Calc_Alpha(): Result {
                return new Result(0);
        }

        /**
         * Calcul de la dérivée de l'angle Alpha entre la surface libre et le fond pour les sections circulaires.
         * @return Dérivée de l'angle Alpha pour une section circulaire, 0 sinon.
         */
        protected Calc_dAlpha(): Result {
                return new Result(0);
        }
}
