import { Nub } from "../nub";
import { ParamCalculability } from "../param";
import { Message } from "../util/message";
import { Result } from "../util/result";
import { ParallelStructureParams } from "./parallel_structure_params";
import { Structure } from "./structure";

/**
 * Calcul de une ou plusieurs structures hydrauliques en parallèles
 * reliées par les cotes amont et aval et dont le débit est égal à la
 * somme des débits de chaque structure.
 */
export class ParallelStructure extends Nub {

    /** Tableau des structures hydrauliques en parallèle */
    public structures: Structure[] = [];

    /**
     * paramètres castés au bon type
     */
    get prms(): ParallelStructureParams {
        return this._prms as ParallelStructureParams;
    }

    /**
     * Mise à jour de Z1 pour toutes les structures en parallèle
     */
    set Z1(Z1: number) {
        this.prms.Z1.v = Z1;
        for (const structure of this.structures) {
            structure.prms.Z1.v = Z1;
            structure.prms.update_h1h2();
        }
    }

    /**
     * Mise à jour de Z2 pour toutes les structures en parallèle
     */
    set Z2(Z2: number) {
        this.prms.Z2.v = Z2;
        for (const structure of this.structures) {
            structure.prms.Z2.v = Z2;
            structure.prms.update_h1h2();
        }
    }

    /**
     * Ajout d'une structure en parallèle
     * @param structure La structure à rajouter
     */
    public addStructure(structure: Structure) {
        this.structures.push(structure);
    }

    /**
     * Supprime une structure hydraulique
     * @param index numéro de la structure dans le tableau
     */
    public deleteStructure(index: number) {
        if (index > -1) {
            this.structures.splice(index, 1);
        } else {
            throw new Error("ParallelStructure.deleteStructure invalid index=" + index);
        }
    }

    /**
     * Calcul du débit des structures en parallèle (sans détail pour chaque structure)
     * @param sVarCalc Variable à calculer (Q uniquement)
     */
    public Equation(sVarCalc: string): Result {
        Structure.CheckEquation(sVarCalc);
        this.Z1 = this.prms.Z1.v;
        this.Z2 = this.prms.Z2.v;
        return this.CalcQ();
    }

    /**
     * Calcul de la somme des débits de chaque structure
     * @param iExcept Index de la structure à ne pas additionner (optionel)
     */
    public CalcQ(iExcept?: number): Result {
        if (iExcept !== undefined) {
            if (iExcept < 0 || iExcept >= this.structures.length) {
                throw new Error(
                    "ParallelStructure.CalcQ iExcept not in [0;" + (this.structures.length - 1) + "]",
                );
            }
        }
        const calcRes: Result = new Result();
        let qTot: number = 0;
        for (let i = 0 ; i < this.structures.length; i++) {
            if (i !== iExcept) {
                const res: Result = this.structures[i].Calc("Q");
                calcRes.addResult(res.result);
                qTot += res.vCalc;
            }
        }
        // Insert le débit total en premier résultat
        calcRes.insertResult(new Result(qTot).result, 0);
        return calcRes;
    }

    /**
     * Calcul du débit total, de la cote amont ou aval ou d'un paramètre d'une structure
     * @param sVarCalc Nom du paramètre à calculer :
     *                 "Q", "Z1", "Z2" ou "n.X" avec "n" l'index de l'ouvrage et "X" son paramètre
     * @param rInit Valeur initiale
     * @param rPrec Précision attendue
     */
    public Calc(sVarCalc: string, rInit?: number, rPrec: number = 0.001): Result {
        switch (sVarCalc) {
            case "Z1" :
            case "Z2" :
            case "Q" :
                return super.Calc(sVarCalc, rInit, rPrec);
            default:
                // Pour les caractéristiques des ouvrages
                return this.CalcStructPrm(sVarCalc, rInit, rPrec);
        }
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        this.prms.Q.calculability = ParamCalculability.EQUATION;
        this.prms.Z1.calculability = ParamCalculability.DICHO;
        this.prms.Z2.calculability = ParamCalculability.DICHO;
    }

    /**
     * Calcul du paramètre d'un des ouvrages en parallèle
     * @param sVarCalc Nom du paramètre à calculer : "n.X" avec "n" l'index de l'ouvrage et "X" son paramètre
     * @param rInit Valeur initiale
     * @param rPrec Précision attendue
     */
    private CalcStructPrm(sVarCalc: string, rInit?: number, rPrec: number = 0.001): Result {
        // Détection de la structure où calculer le paramètre
        let sIndex: string;
        let sPrm: string;
        [sIndex, sPrm] = sVarCalc.split(".");
        const index = parseInt(sIndex, 10);

        // Le débit restant sur la structure en calcul est :
        this.structures[index].prms.Q.v = this.prms.Q.v - this.CalcQ(index).vCalc;

        // Calcul du paramètre de la structure en calcul
        return this.structures[index].Calc(sPrm, rInit, rPrec);
    }
}
