import { Nub } from "../nub";
import { ComputeNodeType, ParamDefinition, ParamDomainValue, ParamsEquation } from "../param";

/**
 * Common parameters of hydraulic structure equations
 */
export class StructureParams extends ParamsEquation {
    /** Débit (m3/s) */
    public Q: ParamDefinition;

    /** Cote de la crête du déversoir ou du radier de la vanne (m) */
    public ZDV: ParamDefinition;

    /** Cote de l'eau amont (m) */
    public Z1: ParamDefinition;

    /** Cote de l'eau aval (m) */
    public Z2: ParamDefinition;

    /** Tirant d'eau amont (m) */
    public h1: ParamDefinition;

    /** Tirant d'eau aval (m) */
    public h2: ParamDefinition;

    /** Ouverture de la vanne
     * @note Pour un seuil cette valeur vaut Infinity
     */
    public W: ParamDefinition;

    /**
     * Paramètres communs à toutes les équations de structure
     * @param rQ Débit (m3/s)
     * @param rZDV Cote de la crête du déversoir ou du radier de la vanne (m)
     * @param rZ1 Cote de l'eau amont (m)
     * @param rZ2 Cote de l'eau aval (m)
     * @param rW Ouverture de vanne (m) (infinity pour un seuil)
     */
    constructor(rQ: number, rZDV: number, rZ1: number, rZ2: number, rW: number = Infinity ) {
        super();
        this.Q = new ParamDefinition(ComputeNodeType.CondDistri, "Q", ParamDomainValue.ANY, rQ);
        this.addParamDefinition(this.Q);
        this.ZDV = new ParamDefinition(ComputeNodeType.CondDistri, "ZDV", ParamDomainValue.ANY, rZDV);
        this.addParamDefinition(this.ZDV);
        this.Z1 = new ParamDefinition(ComputeNodeType.CondDistri, "Z1", ParamDomainValue.ANY, rZ1);
        this.addParamDefinition(this.Z1);
        this.Z2 = new ParamDefinition(ComputeNodeType.CondDistri, "Z2", ParamDomainValue.ANY, rZ2);
        this.addParamDefinition(this.Z2);
        this.h1 = new ParamDefinition(ComputeNodeType.CondDistri, "h1", ParamDomainValue.POS_NULL,
            Math.max(0, this.Z1.v - this.ZDV.v));
        this.addParamDefinition(this.h1);
        this.h2 = new ParamDefinition(ComputeNodeType.CondDistri, "h2", ParamDomainValue.POS_NULL,
            Math.max(0, this.Z2.v - this.ZDV.v));
        this.addParamDefinition(this.h2);
        this.W = new ParamDefinition(ComputeNodeType.CondDistri, "W", ParamDomainValue.POS_NULL, rW);
        this.addParamDefinition(this.W);
    }

    /** Mise à jour des paramètres h1 et h2 à partir de Z1, Z2 et ZDV */
    public update_h1h2() {
        this.h1.v = Math.max(0, this.Z1.v - this.ZDV.v);
        this.h2.v = Math.max(0, this.Z2.v - this.ZDV.v);
    }
}
