import { cLog } from "./log";
import { Message, MessageCode, MessageSeverity } from "./message";
import { ResultElement } from "./resultelement";

/**
 * Résultat global d'un calcul
 * Peut comporter un ou plusieurs Result.
 */
// tslint:disable-next-line:max-classes-per-file
export class Result {
    /**
     * Messages (erreurs, infos, ...)
     */
    private _globalLog: cLog;

    private _results: ResultElement[];

    constructor(v?: number | Message | ResultElement, d?: any) {
        this._globalLog = new cLog();
        this._results = [];
        if (typeof (v) === "number" || v instanceof Message) {
            this._results.push(new ResultElement(v));
        } else if (v instanceof ResultElement) {
            this._results.push(v);
        }

        if (d !== undefined) {
            this.result.extraResults = d;
        }
    }

    /**
     * retourne le journal de la calculette
     */
    public get globalLog() {
        return this._globalLog;
    }

    /**
     * Retourne le résultat du premier ResultElement
     */
    get result(): ResultElement {
        if (this._results.length === 0) {
            throw new Error("appel à la méthode de Result.result() invalide, il n'y a pas au moins un 'ResultElement'");
        }

        return this._results[0];
    }

    /**
     * @return true si il existe au moins 1 ResultElement dans le tableau
     */
    get hasResults(): boolean {
        return this._results.length > 0;
    }

    /**
     * @return le résultat de calcul du 1er ResultElement
     */
    get vCalc(): number {
        return this.result.vCalc;
    }

    /**
     * @return le résultat de calcul du 1er ResultElement
     */
    set vCalc(r: number) {
        this.result.vCalc = r;
    }

    /**
     * retourne le journal du 1er ResultElement
     */
    public get log() {
        // return this._globalLog;
        return this.result.log;
    }

    /**
     * retourne le code du 1er ResultElement
     */
    public get code(): MessageCode {
        return this.result.code;
    }

    /**
     * retourne les résultats complémentaires du 1er ResultElement
     */
    get extraResults() {
        return this.result.extraResults;
    }

    /**
     * Extract one ResultElement as a Result
     * @param i index of the ResultElement
     */
    public extractResult(i: number): Result {
        if (i < 0 || i >= this._results.length) {
            throw new Error("Result.extractResult index outside [0;" + (this._results.length - 1) + "]");
        }
        return new Result(this._results[i]);
    }

    /**
     * Renvoie le nombre de résultats dans le tableau de résultats
     */
    public get nbResults(): number {
        return this._results.length;
    }

    /**
     * Ajoute un ResultElement au tableau
     * @param r ResultElement à ajouter
     */
    public addResult(r: ResultElement) {
        this._results.push(r);
    }

    /**
     * Insert result r at index i
     * @param r Result to insert
     * @param i Index position
     */
    public insertResult(r: ResultElement, i: number) {
        this._results.splice(i, 0, r);
    }

    /**
     * @param name nom du résultat complémentaire
     * @returns le nombre la taille d'un résultat complémentaire
     */
    public extraResultLength(name: string): number {
        const r = this.getExtraResult(name);
        if (r === undefined) {
            return 0;
        }
        return Object.keys(r).length;
    }

    public getExtraResult(name: string): any {
        const res = [];
        for (const r of this._results) {
            const er = r.getExtraResult(name);
            if (er !== undefined) {
                res.push(er);
            }
        }

        switch (res.length) {
            case 0:
                return undefined;

            case 1:
                return res[0];

            default:
                throw new Error(
                    "Result.getExtraResult() : il existe plusieurs ResultElement avec" +
                    " un extratresult dont le nom est '" + name + "'",
                );
        }
    }

    // protected get messageCount(): number {
    //     return this._globalLog.messages.length;
    // }

    /**
     * ajoute un message au journal
     */
    public addMessage(m: Message) {
        this._globalLog.add(m);
    }

    /**
     * ajoute tous les messages d'un journal au journal
     */
    public addLog(l: cLog) {
        this._globalLog.addLog(l);
    }

    /**
     * insert un message en début de liste
     */
    public insertMessage(m: Message) {
        this._globalLog.insert(m);
    }

    /**
     * Teste si tous les résultats + complémentaires sont valides et journal sans erreur
     * @returns true si le test est valide
     */
    public get ok(): boolean {
        return this.resultsOk && !this.hasErrorMessages;
    }

    /**
     * Teste si tous les résultats + complémentaires sont valides
     * @returns true si le test est valide
     */
    public get resultsOk(): boolean {
        if (this._results.length === 0) {
            return false;
        }
        for (const r of this._results) {
            if (!r.resultsOk) {
                return false;
            }
        }
        return true;
    }

    /**
     * @returns true si au moins un message de log comporte un code d'erreur
     */
    public get hasErrorMessages(): boolean {
        // Test du journal global
        for (const m of this._globalLog.messages) {
            if (m.getSeverity() === MessageSeverity.ERROR) {
                return true;
            }
        }
        // Test des journaux des résultats
        for (const r of this._results) {
            if (r.hasErrorMessages) {
                return true;
            }
        }

        return false;
    }
}
