import { cLog } from "./log";
import { Message, MessageCode, MessageSeverity } from "./message";

/**
 * Résultat de calcul comprenant la valeur du résultat et des calculs annexes (flag, calculs intermédiaires...)
 * Représente une valeur unique
 */
// tslint:disable-next-line:max-classes-per-file
export class ResultElement {
    /**
     * Valeur calculée
     */
    private _vCalc: number;

    /**
     * Messages (erreurs, infos, ...)
     */
    private _log: cLog;

    /**
     * résultats complémentaires
     */
    private _extraResults: { [key: string]: any };

    constructor(v?: number | Message) {
        this._log = new cLog();
        if (typeof v === "number") {
            this._vCalc = v;
        } else if (v instanceof Message) {
            this.addMessage(v);
        }
        this._extraResults = {};
    }

    get vCalc() { return this._vCalc; }

    set vCalc(r: number) {
        this._vCalc = r;
    }

    public get log() {
        return this._log;
    }

    get code(): MessageCode {
        if (this.messageCount !== 1) {
            throw new Error("appel à Result.code invalide, il n'y pas exactement un message de log");
        }
        return this._log.messages[0].code;
    }

    get extraVar() {
        if (this.messageCount !== 1) {
            throw new Error("appel à Result.extraVar invalide, il n'y pas exactement un message de log");
        }
        return this._log.messages[0].extraVar;
    }

    private get hasExtraResults(): boolean {
        return Object.keys(this._extraResults).length > 0;
    }

    public get resultsOk(): boolean {
        return this._vCalc !== undefined || this.hasExtraResults;
    }

    private get messageCount(): number {
        return this._log.messages.length;
    }

    public get hasErrorMessages(): boolean {
        for (const m of this._log.messages) {
            if (m.code !== MessageCode.ERROR_OK && m.getSeverity() === MessageSeverity.ERROR) {
                return true;
            }
        }
        return false;
    }

    public get ok(): boolean {
        return this.resultsOk && !this.hasErrorMessages;
    }

    /**
     * ajoute un message au journal
     */
    public addMessage(m: Message) {
        this._log.add(m);
    }

    /**
     * insert un message en début de liste
     */
    public insertMessage(m: Message) {
        this._log.insert(m);
    }

    /**
     * résultats complémentaires
     */

    get extraResults() {
        return this._extraResults;
    }

    set extraResults(d: { [key: string]: any }) {
        this._extraResults = d;
    }

    public addExtraResult(name: string, value: any) {
        this._extraResults[name] = value;
    }

    /**
     * @param name nom du résultat complémentaire à retourner
     * @returns un résultat complémentaire par son nom
     */
    public getExtraResult(name: string): any {
        for (const n in this._extraResults) {
            if (n === name) {
                return this._extraResults[n];
            }
        }
        return undefined;
    }

    public toString(): string {
        if (this._vCalc !== undefined) {
            return String(this._vCalc);
        }
        return JSON.stringify(this);
    }
}
